/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * JBIListener.java - Initial Developper : ofabre
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.xslt.listener;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;
import java.util.logging.Level;

import javax.activation.DataHandler;
import javax.jbi.messaging.MessagingException;
import javax.security.auth.Subject;
import javax.xml.transform.Source;
import javax.xml.transform.Templates;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.URIResolver;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.performance.PerformanceNotification;
import org.ow2.petals.component.framework.util.StringHelper;
import org.ow2.petals.se.xslt.Constants;
import org.ow2.petals.se.xslt.ServiceUnitURIResolver;
import org.ow2.petals.se.xslt.StreamDataSource;
import org.ow2.petals.se.xslt.Xslt;
import org.ow2.petals.se.xslt.XsltImpl;

/**
 * The XSLT binding component JBIListener
 * 
 * @author alouis, ofabre, cgirodengo
 * 
 */
public class JBIListener extends AbstractJBIListener {

    private static Map<String, Templates> TEMPLATES_CACHE = new Hashtable<String, Templates>();

    private Xslt xslt;

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.petals.component.framework.listener.AbstractListener#init()
     */
    @Override
    protected void init() {
        this.xslt = new XsltImpl();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.petals.component.framework.listener.AbstractJBIListener
     * #onJBIMessage(org.objectweb.petals.component.framework.util.Exchange)
     */
    @Override
    protected void onJBIMessage(final Exchange exchange) {
        this.getLogger().log(Level.INFO, "A new exchange is received.");

        if (exchange.isInOutPattern()) {
            this.processExchange(exchange);
        } else {
            // refuse the others patterns than InOut
            this.getLogger().warning(
                    "Exchange pattern (" + exchange.getPattern() + ") not supported");
            exchange.setError(new Exception("Unsupported exchange pattern"));
        }
    }

    /**
     * Transform the outputstream as an IN DataHandler
     * 
     * @param outputStream
     * @param attachmentName
     * @return
     */
    private DataHandler createDataSourceFromOutputStream(final ByteArrayOutputStream outputStream,
            final String attachmentName) {
        this.getLogger().log(Level.INFO, "Put the result as an attachment.");
        final ByteArrayInputStream input = new ByteArrayInputStream(outputStream.toByteArray());
        final StreamDataSource datasource = new StreamDataSource(input, attachmentName);
        return new DataHandler(datasource);
    }

    /**
     * Transform the Outputstream as a Source
     * 
     * @param outputStream
     * @return
     */
    private StreamSource createSourceFromOutputStream(final ByteArrayOutputStream outputStream) {
        this.getLogger().log(Level.INFO, "Put the result as Source.");

        final StreamSource sourceResult = new StreamSource();
        final ByteArrayInputStream result = new ByteArrayInputStream(outputStream.toByteArray());
        sourceResult.setInputStream(result);
        return sourceResult;
    }

    /**
     * Retrieve the Templates associated to this XSL, and update the
     * corresponding entry in the cache. Identifier is composed of a String
     * which value is the location of the XSL file.
     * 
     * @param xslResourceInputStream
     * @param identifier
     * @return
     * @throws TransformerConfigurationException
     *      Failed to create the XSL template
     */
    private Templates getTemplatesAndUpdateCache(final InputStream xslResourceInputStream,
            final String identifier) throws TransformerConfigurationException {
        Templates result = null;

        if (xslResourceInputStream == null) {
            return null;
        }

        final Source source = new StreamSource(xslResourceInputStream);

        final TransformerFactory transformerFactory = TransformerFactory.newInstance();
        result = transformerFactory.newTemplates(source);

        TEMPLATES_CACHE.put(identifier, result);

        this.getLogger().log(Level.INFO, "XSL stored in cache. Identifier:" + identifier);
        return result;
    }

    /**
     * Retrieve the InputStream associated to the desired xsl from the Consumes.
     * 
     * TODO : The templates must be loaded and put in the cache during the start
     * of the SU and removed during the shutdown.
     * 
     * @return
     * @throws FileNotFoundException
     * @throws TransformerConfigurationException
     */
    private Templates getXslTemplates() {
        this.getLogger().log(Level.INFO, "Retrieve the XSL style sheet to use.");
        final String xslResourcePath = this.getExtensions().get(Constants.XSL_FILE);

        Templates xslTemplate = JBIListener.TEMPLATES_CACHE.get(this.getInstallRoot()
                + xslResourcePath);

        if (xslTemplate == null) {
            this.getLogger().info("Getting XSD from resource : " + xslResourcePath);

            InputStream xslStream;
            // try to get the XSL from the filesystem
            try {
                xslStream = new FileInputStream(new File(this.getInstallRoot(), xslResourcePath));
            } catch (FileNotFoundException e) {
                throw new RuntimeException("Failed to retrieve XSL resource - "
                        + e.getClass().getSimpleName() + " - " + e.getMessage());
            }

            try {
                xslTemplate = this.getTemplatesAndUpdateCache(xslStream, this.getInstallRoot()
                        + xslResourcePath);
            } catch (TransformerConfigurationException e) {
                throw new RuntimeException("Failed to build XSL templates - "
                        + e.getClass().getSimpleName() + " - " + e.getMessage());
            }

        }
        return xslTemplate;
    }

    /**
     * Process the transformation and set the response as Source or attachment
     * 
     * @param exchange
     * @throws BusinessException
     * @throws MessagingException
     */
    @SuppressWarnings("deprecation")
    private void processExchange(final Exchange exchange) {
        this.getLogger().log(Level.INFO, "Process the transformation.");

        try {
            // Get in Message content
            Source inContent = exchange.getInMessageContentAsSource();

            Subject subject = exchange.getInMessage().getSecuritySubject();

            PerformanceNotification performanceNotification = this.component
                    .createPerformanceNotificationSendingToOutside("Time for XSLT transformation",
                            exchange);

            // get the XSL template for current Service Unit
            final Templates xslTransformers = this.getXslTemplates();
            if (xslTransformers == null) {
                exchange.setError(new RuntimeException(
                        "No XSL style sheet is found to transform the document."));
            }

            // transform the input
            final ByteArrayOutputStream outContent = this.xslt
                    .transform(inContent, xslTransformers, this.getInstallRoot());

            final String outputAttachmentName = this.getExtensions().get(
                    Constants.XSL_OUTPUT_ATTACHMENT_NAME);
            if (StringHelper.isNullOrEmpty(outputAttachmentName)) {
                // set the response as output Source
                final Source source = this.createSourceFromOutputStream(outContent);
                exchange.setOutMessageContent(source, subject);
            } else {
                // set the response as attachment
                final DataHandler attachment = this.createDataSourceFromOutputStream(outContent,
                        outputAttachmentName);
                final HashMap<String, DataHandler> attachments = new HashMap<String, DataHandler>();
                attachments.put(outputAttachmentName, attachment);
                exchange.setOutMessageAttachments(attachments);
                final String outSource = "<attached-files><file-name>" + outputAttachmentName
                        + "</file-name></attached-files>";
                exchange.setOutMessageContent(outSource, subject);
            }

            this.component.sendPerformanceNotification(performanceNotification);

        } catch (Exception e) {
            if (e instanceof RuntimeException) {
                throw (RuntimeException) e;
            } else {
                throw new RuntimeException(e.getClass().getSimpleName() + " - " + e.getMessage());
            }
        }
    }
}
